/*
** prpn.c: A RPN calculator
**
** Copyright (c) 2003 Peter Eriksson <pen@lysator.liu.se>
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>


char *argv0 = "prpn";
double *stack = NULL;
int stacksize = 0;
int stackpos = 0;
char suffixlist[] = " kMGTPEZY";
int sum_f = 0;
int nan_f = 0;
char t_suffix = 0;
int dotprec = -1;


void
stack_push(double val)
{
    if (stackpos >= stacksize)
    {
	stacksize += 256;
	if (!stack)
	    stack = malloc(sizeof(double)*stacksize);
	else
	    stack = realloc(stack, sizeof(double)*stacksize);
	if (!stack)
	{
	    fprintf(stderr, "%s: stack_push: allocation failed: %s\n",
		    argv0, strerror(errno));
	    exit(1);
	}
    }

    stack[stackpos++] = val;
}

double
stack_pop(void)
{
    double val;

    
    if (stackpos > 0)
    {
	val = stack[--stackpos];
	return val;
    }

    fputs("> ", stdout);
    fflush(stdout);
    if (scanf("%lf", &val) == 1)
	return val;

    fprintf(stderr, "%s: POP: %s\n", argv0, strerror(errno));
    exit(1);
}




int
suffix2exp(int suffix)
{
  int i;

  if (!suffix)
    return 0;
  
  for (i = 0; suffixlist[i] && suffix != suffixlist[i]; i++)
    ;

  if (suffixlist[i])
    return i*3;

  return -1;
}


void
print_value(double val)
{
    int j;

    
    for (j = 0; (t_suffix || fabs(val / 1000) >= 1) &&
	     suffixlist[j] && suffixlist[j] != t_suffix; j++)
    {
	val /= 1000;
    }
    
    
    if (dotprec >= 0)
    {
	printf("%.*lf", dotprec, val);
	if (j > 0)
	    putchar(suffixlist[j]);
    }
    else
    {
	printf("%lf", val);
	if (j > 0)
	    putchar(suffixlist[j]);
    }
    putchar('\n');
}


int
token_eval(char *token)
{
    double val;
    char suffix = 0;
    int expval = 0;
    
    
    if (sscanf(token, "%lf%c", &val, &suffix) >= 1)
    {
	if (nan_f && isnan(val))
	    val = 0;
	
	if (suffix)
	    expval = suffix2exp(suffix);
	else
	    expval = 0;
	
	if (expval < 0)
	{
	    fprintf(stderr, "%s: invalid suffix: %s (%d)\n",
		    argv0, token, suffix);
	    exit(1);
	}
	
	if (expval)
	    val *= pow(10.0, expval);
	
	stack_push(val);
    }
    
    else if (strcmp(token, "+") == 0)
    {
	double v1, v2;
	
	v1 = stack_pop();
	v2 = stack_pop();
	stack_push(v2+v1);
    }
    
    else if (strcmp(token, "-") == 0)
    {
	double v1, v2;
	
	v1 = stack_pop();
	v2 = stack_pop();
	stack_push(v2-v1);
    }
    
    else if (strcmp(token, "*") == 0)
    {
	double v1, v2;
	
	v1 = stack_pop();
	v2 = stack_pop();
	stack_push(v2*v1);
    }
    
    else if (strcmp(token, "/") == 0)
    {
	double v1, v2;
	
	v1 = stack_pop();
	v2 = stack_pop();
	stack_push(v2/v1);
    }
    
    else if (strcmp(token, "neg") == 0)
    {
	double v1;
	
	v1 = stack_pop();
	stack_push(-v1);
    }
    
    else if (strcmp(token, "rnd") == 0)
    {
	double v1 = drand48();

	stack_push(v1);
    }

    else if (strcmp(token, "ln") == 0)
    {
	double v1 = stack_pop();

	stack_push(log(v1));
    }

    else if (strcmp(token, "log") == 0)
    {
	double v1 = stack_pop();

	stack_push(log10(v1));
    }

    else if (strcmp(token, "cos") == 0)
    {
	double v1 = stack_pop();

	stack_push(cos(v1));
    }

    else if (strcmp(token, "cosh") == 0)
    {
	double v1 = stack_pop();

	stack_push(cosh(v1));
    }

    else if (strcmp(token, "acos") == 0)
    {
	double v1 = stack_pop();

	stack_push(acos(v1));
    }

    else if (strcmp(token, "asin") == 0)
    {
	double v1 = stack_pop();

	stack_push(asin(v1));
    }

    else if (strcmp(token, "atan") == 0)
    {
	double v1 = stack_pop();

	stack_push(atan(v1));
    }

    else if (strcmp(token, "ceil") == 0)
    {
	double v1 = stack_pop();

	stack_push(ceil(v1));
    }


    else if (strcmp(token, "cosh") == 0)
    {
	double v1 = stack_pop();

	stack_push(cosh(v1));
    }

    else if (strcmp(token, "exp") == 0)
    {
	double v1 = stack_pop();

	stack_push(exp(v1));
    }
	
    else if (strcmp(token, "abs") == 0)
    {
	double v1 = stack_pop();

	stack_push(fabs(v1));
    }
	
    else if (strcmp(token, "floor") == 0)
    {
	double v1 = stack_pop();

	stack_push(floor(v1));
    }
	
    else if (strcmp(token, "mod") == 0)
    {
	double v1 = stack_pop();
	double v2 = stack_pop();

	stack_push(fmod(v2, v1));
    }
    
    else if (strcmp(token, "pow") == 0)
    {
	double v1 = stack_pop();
	double v2 = stack_pop();

	stack_push(pow(v2, v1));
    }

    else if (strcmp(token, "sin") == 0)
    {
	double v1 = stack_pop();

	stack_push(sin(v1));
    }

    else if (strcmp(token, "sinh") == 0)
    {
	double v1 = stack_pop();

	stack_push(sinh(v1));
    }

    else if (strcmp(token, "sqrt") == 0)
    {
	double v1 = stack_pop();

	stack_push(sqrt(v1));
    }

    else if (strcmp(token, "tan") == 0)
    {
	double v1 = stack_pop();

	stack_push(tan(v1));
    }

    else if (strcmp(token, "tanh") == 0)
    {
	double v1 = stack_pop();

	stack_push(tanh(v1));
    }

    else if (strcasecmp(token, "isnan") == 0)
    {
	double v1 = stack_pop();

	stack_push(isnan(v1));
    }

    /* Constants... */
    
    else if (strcasecmp(token, "e") == 0)
    {
	stack_push(M_E);
    }

    else if (strcasecmp(token, "pi") == 0)
    {
	stack_push(M_PI);
    }
    
    

    
    else if (strcasecmp(token, "dup") == 0)
    {
	double v1 = stack_pop();
	stack_push(v1);
	stack_push(v1);
    }
    
    else if (strcasecmp(token, "drop") == 0)
    {
	(void) stack_pop();
    }
    
    else if (strcasecmp(token, "depth") == 0)
    {
	stack_push((double) stackpos);
    }
    
    else if (strcasecmp(token, "swap") == 0)
    {
	double v1, v2;
	
	v1 = stack_pop();
	v2 = stack_pop();
	stack_push(v1);
	stack_push(v2);
    }
    
    else if (strcmp(token, ".") == 0)
    {
	double v1;
	
	v1 = stack_pop();
	print_value(v1);
    }
    
    else if (strcmp(token, "exit") == 0)
    {
	double v1;
	
	v1 = stack_pop();
	exit((int) v1);
    }
    
    else if (strcmp(token, "prec") == 0)
    {
	double v1;
	
	v1 = stack_pop();
	dotprec = (int) v1;
    }

    else
    {
	fprintf(stderr, "%s: Invalid number/token: %s\n",
		argv0, token);
	return -1;
    }

    return 0;
}


char *
token_get(void)
{
    static char buf[1024];

    if (scanf("%s", buf) == 1)
	return buf;

    return NULL;
}


int
main(int argc,
     char *argv[])
{
    int i, j, len;
    double val, tval;
    char *token = NULL;
    

    if (argv[0])
	argv0 = argv[0];
    
    tval = 0;
    
    for (i = 1; i < argc && argv[i][0] == '-'; ++i)
	switch (argv[i][1])
	{
	  case '-':
	    ++i;
	    goto Values;

	  case 'n':
	    nan_f++;
	    break;
	    
	  case 's':
	    sum_f++;
	    break;

	  case 'S':
	    t_suffix = argv[i][2];
	    if (!t_suffix)
	      t_suffix = ' ';
	    else if (t_suffix == '-')
	      t_suffix = 0;
	    break;
	    
	  case 'p':
	    if (sscanf(argv[i]+2, "%u", &dotprec) != 1)
	    {
		fprintf(stderr, "%s: invalid precision: %s\n",
			argv0, argv[i]);
		exit(1);
	    }
	    break;

	  default:
	    fprintf(stderr, "%s: invalid switch: %s\n",
		    argv0, argv[i]);
	    exit(1);
	}

  Values:
    if (i < argc)
    {
	for (;i < argc; ++i)
	{
	    token = argv[i];
	    
	    token_eval(token);
	}
    }
    else
    {
	while (token = token_get())
	    token_eval(token);
    }

    for (i = 0; i < stackpos; i++)
    {
	val = stack[i];

	if (sum_f)
	{
	    tval += val;
	    if (i+1 < stackpos)
		continue;
	}
	else
	    tval = val;
    
	print_value(tval);
    }
    
    return 0;
}
